/*
 * iterativeUpdates.c
 *
 *  Created on: Oct 1, 2014
 *      Author: smetzler
 */
#include <stdio.h>
#include <math.h>
#include <omp.h>
#include "btc.h"
//#include <simstruc.h> // needed to forward printf output to matlab

/**
 * Calculates the outer product of two binary vectors.
 */
uint64_t* outerProduct(const uint64_t* A, const uint64_t* B, const dim2 dim) {
	const uint64_t integerPerColumn = numberOfUint64s(dim.x);
	uint64_t* result = checkedCalloc(integerPerColumn*dim.y);

	for (uint64_t j = 0; j < dim.y; j++) {
		if (B[j/64] & (ONE << j%64)) {
			for (uint64_t i = 0; i < integerPerColumn; i++) {
				result[j*integerPerColumn + i] = A[i];
			}
		}
	}

	return result;
}

/**
 * Turns a matrix into rank-1.
 */
void toRank1(uint64_t* mat, const uint64_t* colsums, const dim2 dim) {
	const int integerPerColumn = numberOfUint64s(dim.x);
	decom rank1 = rank1approx_parallel(mat,colsums,dim);
	uint64_t* centr = outerProduct(rank1.A, rank1.B, dim);
	memcpy(mat, centr, integerPerColumn*dim.y*sizeof(uint64_t));
	free(centr);
	free(rank1.A);
	free(rank1.B);
}

/**
 * Clusters the same way as btc() but does not decompose the centroids to rank 1.
 */
clustering getClusteringWithFullCentroids(const bitwiseTensor* X, const dim3 dim, const uint64_t numberOfClusters,
		                                  const uint64_t reps, const uint64_t numThreads) {
	const uint64_t integerPerColumn = numberOfUint64s(dim.x);
	const uint64_t integerPerTube = numberOfUint64s(dim.z);
	if (numThreads != 0)
		omp_set_num_threads(numThreads);
	clustering result = {.centroids = checkedCalloc(integerPerColumn*dim.y*numberOfClusters),
			.assignment = checkedCalloc(integerPerTube*numberOfClusters), .similarity = 0};

	for (uint64_t rep = 0; rep < reps; rep++) {
		// select numberOfClusters x*y slices of X and assign each to its closest centroid
		uint64_t* permutedZ = randomPermutation(dim.z);
#ifdef TESTMODE
		printf("Warning: non random choice of cluster centroids.\n");
		permutedZ = range(dim.z, rep);
#endif
		uint64_t* centroids[numberOfClusters];
		uint64_t* d = checkedCalloc(dim.z * numberOfClusters); // distances to centroids
		for (uint64_t k = 0; k < numberOfClusters; k++) {
			const uint64_t slice = permutedZ[k];
			centroids[k] = &(X->tensor[slice * dim.y * integerPerColumn]);
			for (uint64_t r = 0; r < dim.z; r++) {
				for (uint64_t j = 0; j < dim.y; j++) {
					for (uint64_t i = 0; i < integerPerColumn; i++) {
						d[k * dim.z + r] +=	POPCNT64(X->tensor[r*dim.y*integerPerColumn + integerPerColumn*j + i]
										^ *(centroids[k] + integerPerColumn*j + i));
					}
				}
			}
		}
		free(permutedZ);

		int64_t simi = 0;
		uint64_t* clusterLabels = checkedCalloc(dim.z);
#pragma omp parallel for reduction(+:simi)
		for (uint64_t r = 0; r < dim.z; r++) {
			uint64_t minD = UINT64_MAX;
			uint64_t indexOfMinD = INT_MAX;
			for (uint64_t k = 0; k < numberOfClusters; k++) {
				if (d[k * dim.z + r] < minD) {
					minD = d[k * dim.z + r];
					indexOfMinD = k;
				}
			}
			simi += (dim.x * dim.y - minD);
			clusterLabels[r] = indexOfMinD;
		}
		free(d);

		if (simi > result.similarity) {
			result.similarity = simi;
			for (uint64_t k = 0; k < numberOfClusters; k++) {
				memcpy(&result.centroids[integerPerColumn * dim.y * k], centroids[k],
						integerPerColumn*dim.y*sizeof(uint64_t));
				for (uint64_t r = 0; r < dim.z; r++) {
					if (clusterLabels[r] == k) {
						result.assignment[integerPerTube * k + r / 64] |= ONE << r % 64;
					}
				}
			}
		}
		free(clusterLabels);
	}

	return result;
}

/**
 * Gets an initial cluster assignment. Optionally, the centroids can be restricted to rank-1.
 */
clustering getInitialClustering(const bitwiseTensor* X, const dim3 dim, const uint64_t numberOfClusters,
		                        const uint64_t reps, const uint64_t numThreads, const bool rank1Centroids) {
	if (rank1Centroids) {
		const uint64_t integerPerColumn = numberOfUint64s(dim.x);
		const uint64_t integerPerTube = numberOfUint64s(dim.z);

		clustering result = {.centroids = checkedCalloc(integerPerColumn*dim.y*numberOfClusters),
				.assignment = checkedCalloc(integerPerTube*numberOfClusters), .similarity = 0};
		decom3 btcRes = btc(X, dim, numberOfClusters, reps, numThreads);
		for (uint64_t k = 0; k < numberOfClusters; k++) {
			uint64_t* centr = outerProduct(&btcRes.A[integerPerColumn * k],
					&btcRes.B[numberOfUint64s(dim.y) * k], (dim2) {dim.x, dim.y});
			memcpy(&result.centroids[integerPerColumn * dim.y * k], centr,
					integerPerColumn*dim.y*sizeof(uint64_t));
			free(centr);
		}
		memcpy(result.assignment, btcRes.C, integerPerTube*numberOfClusters*sizeof(uint64_t));
		result.similarity = btcRes.simi;
		free(btcRes.A);
		free(btcRes.B);
		free(btcRes.C);
		return result;
	} else {
		return getClusteringWithFullCentroids(X,dim,numberOfClusters,reps,numThreads);
	}
}

/**
 * Calculates the majority element for the members of one cluster by binary addition. Optionally returns
 * the rank-1 decomposition of the obtained majority matrix.
 */
uint64_t* calculateMajorities(uint64_t** slicesOfCluster, const uint64_t numberOfSlices, const dim2 dim, const bool rank1Centroid) {
	const uint64_t integerPerColumn = numberOfUint64s(dim.x);
	const uint64_t digitsInSum = 1 + (numberOfSlices > 0 ? ceil(log2(numberOfSlices)) : 0);

	uint64_t* majorities = checkedCalloc(integerPerColumn*dim.y);
	uint64_t* colsums = checkedCalloc(dim.y);

#pragma omp parallel for
	for (uint64_t j = 0; j < dim.y; j++) {
		uint64_t* binarySum = checkedCalloc(integerPerColumn * digitsInSum);
		for (uint64_t s = 0; s < numberOfSlices; s++) {
			uint64_t* parity1 = checkedCalloc(integerPerColumn);
			uint64_t* parity2 = checkedCalloc(integerPerColumn);
			memcpy(parity1, (slicesOfCluster[s] + j*integerPerColumn), integerPerColumn*sizeof(uint64_t));

			for (uint64_t d = 0; d < digitsInSum; d++) {
				for (uint64_t i = 0; i < integerPerColumn; i++) {
					parity2[i] = parity1[i] & binarySum[d*integerPerColumn + i];
					binarySum[d*integerPerColumn + i] ^= parity1[i];
				}
				memcpy(parity1, parity2, integerPerColumn*sizeof(uint64_t));
			}
			free(parity1);
			free(parity2);
		}

		uint64_t* sum = checkedCalloc(dim.x);
		for (uint64_t d = 0; d < digitsInSum; d++) {
			for (uint64_t i = 0; i < dim.x; i++) {
				if (binarySum[d*integerPerColumn + i/64] & ONE << i % 64) {
					sum[i] |= ONE << d;
				}
			}
		}
		for (uint64_t i = 0; i < dim.x; i++) {
			if (2 * sum[i] >= numberOfSlices) {
				//printf("j = %lu\ti = %i\tsum = %lu/%lu\n", j,i, sum[i],numberOfSlices);
				majorities[integerPerColumn*j + i/64] |= ONE << i % 64;
				colsums[j]++;
			}
		}
		free(binarySum);
		free(sum);
	}

	if (rank1Centroid) {
		toRank1(majorities,colsums,dim);
	}

	free(colsums);

	return majorities;
}

/**
 * naive summation to obtain majorities
 */
uint64_t* naiveMajorities(uint64_t** slicesOfCluster, const uint64_t numberOfSlices, const dim2 dim, bool rank1Centroid) {
	const uint64_t integerPerColumn = numberOfUint64s(dim.x);
	uint64_t* newCentroid = checkedCalloc(integerPerColumn * dim.y);
	uint64_t* counters = checkedCalloc(dim.x*dim.y);

	for (uint64_t s = 0; s < numberOfSlices; s++) {
		for (uint64_t j = 0; j < dim.y; j++) {
			for (uint64_t i = 0; i < dim.x; i++) {
				if (*(slicesOfCluster[s] + j*integerPerColumn + i/64) & (ONE << i%64)) {
					counters[i + j*dim.x] += ONE;
				}
			}
		}
	}

	uint64_t* colsums = checkedCalloc(dim.y);
	for (uint64_t j = 0; j < dim.y; j++) {
		for (uint64_t i = 0; i < dim.x; i++) {
			if (counters[i + j*dim.x] * 2 >= numberOfSlices) {
				//printf("j = %lu\ti = %lu\tsum = %lu/%lu\n", j,i, counters[i + j*x],numberOfSlices);
				newCentroid[j*integerPerColumn + i/64] |= (ONE << i%64);
				colsums[j]++;
			}
		}
	}

	if (rank1Centroid) {
		toRank1(newCentroid,colsums,dim);
	}

	free(counters);
	free(colsums);

	return newCentroid;
}

/**
 * For each cluster, this function calculates the majority element in every position,
 * hence the new centroid. Optionally, this centroid can be decomposed into rank-1.
 *
 * The new centroids are stored in clustering* c.
 */
void getNewCentroids(const bitwiseTensor* X, clustering* c, const dim3 dim, const uint64_t numberOfClusters, const bool rank1Centroid) {
	const uint64_t integerPerColumn = numberOfUint64s(dim.x);
	const uint64_t integerPerTube = numberOfUint64s(dim.z);

	for (uint64_t k = 0; k < numberOfClusters; k++) {
		uint64_t numberOfSlices = 0;
		uint64_t* slicesOfCluster[dim.z];
		for (uint64_t r = 0; r < dim.z; r++) {
			if (c->assignment[integerPerTube*k + r/64] & ONE << r % 64) {
				slicesOfCluster[numberOfSlices] = &(X->tensor[r*dim.y*integerPerColumn]);
				numberOfSlices++;
			}
		}
		uint64_t* newCentroid = calculateMajorities(slicesOfCluster,numberOfSlices, (dim2) {dim.x,dim.y} ,rank1Centroid);
		memcpy(&c->centroids[k * integerPerColumn * dim.y], newCentroid, integerPerColumn*dim.y*sizeof(uint64_t));
		free(newCentroid);
	}
}

/**
 * Calculates new clustering assignment.
 *
 * The new assignment is stored in clustering* c
 */
void getNewAssignment(const bitwiseTensor* X, clustering* c, const dim3 dim, const uint64_t numberOfClusters) {
	const uint64_t integerPerColumn = numberOfUint64s(dim.x);
	const uint64_t integerPerTube = numberOfUint64s(dim.z);

	uint64_t* d = checkedCalloc(dim.z*numberOfClusters); // distances to centroids
	memset(c->assignment, 0, sizeof(uint64_t)*numberOfClusters*integerPerTube); // reset old assignment
#pragma omp parallel for
	for (uint64_t k = 0; k < numberOfClusters; k++) {
		for (uint64_t r = 0; r < dim.z; r++) {
			for (uint64_t j = 0; j < dim.y; j++) {
				for (uint64_t i = 0; i < integerPerColumn; i++) {
					d[k*dim.z+r] += POPCNT64(X->tensor[r*dim.y*integerPerColumn + integerPerColumn*j + i]
									   ^ c->centroids[k*dim.y*integerPerColumn + integerPerColumn*j + i]);
				}
			}
		}
	}

	int64_t simi = 0;
	uint64_t* clusterLabels = checkedCalloc(dim.z);
#pragma omp parallel for reduction(+:simi)
	for (uint64_t r = 0; r < dim.z; r++) {
		uint64_t minD = UINT64_MAX;
		uint64_t indexOfMinD = INT_MAX;
		for (uint64_t k = 0; k < numberOfClusters; k++) {
			if (d[k*dim.z+r] < minD) {
				minD = d[k*dim.z+r];
				indexOfMinD = k;
			}
		}
		simi += (dim.x*dim.y - minD);
		clusterLabels[r] = indexOfMinD;
	}
	free(d);
	c->similarity = simi;
	for (uint64_t k = 0; k < numberOfClusters; k++) {
		for (uint64_t r = 0; r < dim.z; r++) {
			if (clusterLabels[r] == k) {
				c->assignment[integerPerTube*k + r/64] |= ONE << r % 64;
			}
		}
	}
	free(clusterLabels);
}

/**
 * Runs the clustering updates.
 */
void iterativeUpdates(const bitwiseTensor* X, clustering* clustering, const dim3 dim, uint64_t numberOfClusters, bool rank1) {
	uint64_t oldSimilarity = 0;
	int it = 0;
	while (it < MAX_ITERATIONS) {
		if (oldSimilarity == clustering->similarity) {
			break;
		}
		oldSimilarity = clustering->similarity;

		getNewCentroids(X,clustering,dim,numberOfClusters,rank1);
		getNewAssignment(X,clustering,dim,numberOfClusters);

		it++;
		printf("%lu, %lu, improvement: %ld\n", oldSimilarity, clustering->similarity, (int64_t) clustering->similarity - oldSimilarity);
	}
}
