#include "mex.h"
#include "btc.h"
#include <simstruc.h> // needed to forward printf output to matlab

/* Given a sparse binary 3-way tensor, this function provides a boolean tensor decomposition.
 * This means, it clusters the the first mode and gives a rank-1 decomposition for each
 * cluster centroid. The required input comprises the sparse tensor, i.e. a list of coordinates
 * of the 1's, its dimensionality, the number of clusters to find in the third mode and,
 * as the cluster centroids are found by a randomized algorithm, the number of repetitions out
 * of which the best solution is chosen. Optionally, the number of threads to use for the
 * calculation can be specified. If not specified, the system-dependent OpenMP default number
 * is used.
 * The output are three binary matrices, the first two are provided as sparse, the last one of
 * them being a cluster assignment matrix is dense, and a scalar indicating the similarity
 * between the original tensor and the found approximation.
 *
 * For an input tensor X of dimensionality x*y*z and k clusters, the output will be of x*k,
 * y*k and z*k. Its Khatri-Rao product again forms an x*y*z tensor X', an approximation of X.
 *
 * This function is intended to be called from Matlab.
 */
void mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[]) {
    if (!(nrhs == 5 || nrhs == 4))
        mexErrMsgTxt("Wrong number of input arguments. Four to five arguments expected.");
    if (nlhs != 4)
        mexErrMsgTxt("Wrong number of output arguments. Exactly four arguments expected.");

	const double const *mdim = mxGetData(prhs[1]);
    const dim3 dim = { .x = mdim[0], .y = mdim[1], .z = mdim[2]};
    const mwSize k = mxGetScalar(prhs[2]);
    const mwSize repetitions = mxGetScalar(prhs[3]);
    const mwSize numThreads = (nrhs == 5) ? mxGetScalar(prhs[4]) : 0;
    const size_t m = mxGetM(prhs[0]);
    double* coordinates = mxGetData(prhs[0]);

    /* Convert to bitwise representation and get decomposition. */
    bitwiseTensor* bitPl = sparseToBitRepresentation(coordinates, m, dim);
    decom3 result = btc(bitPl,dim,k,repetitions,numThreads);

	free(bitPl->tensor);
    free(bitPl->onesPerColumn);
    free(bitPl);

    /* Prepare output. */
    arrayWithLength* A;
    A= toSparseMatlabRepresentation(result.A,(dim3){dim.x,k,1});
    plhs[0] = mxCreateDoubleMatrix(A->length/3, 2, mxREAL);
    double* pA = mxGetPr(plhs[0]);
    memcpy(pA, A->data, (A->length - A->length/3)*sizeof(double));

    arrayWithLength* B;
    B= toSparseMatlabRepresentation(result.B,(dim3){dim.y,k,1});
    plhs[1] = mxCreateDoubleMatrix(B->length/3, 2, mxREAL);
    double* pB = mxGetPr(plhs[1]);
    memcpy(pB, B->data, (B->length - B->length/3)*sizeof(double));

    arrayWithLength* C;
    C= toSparseMatlabRepresentation(result.C,(dim3){dim.z,k,1});
    plhs[2] = mxCreateDoubleMatrix(C->length/3, 2, mxREAL);
    double* pC = mxGetPr(plhs[2]);
    memcpy(pC, C->data, (C->length - C->length/3)*sizeof(double));

    plhs[3] = mxCreateDoubleScalar(result.simi);
    return;
}
