//============================================================================
// Name        : matlabToBitwise.c
// Author      : Saskia Metzler
// Version     :
// Copyright   :
// Description :
//============================================================================

#include <errno.h>
#include <string.h>
#include <time.h>
#include <stdlib.h>
#include <stdio.h>
#include <x86intrin.h>
#include "btc.h"
//#include <simstruc.h> // needed to forward printf output to matlab

#ifndef POPCNT64
#define POPCNT64(x) __builtin_popcountll(x)
#endif

uint64_t numberOfUint64s(const uint64_t n) {
	return n / 64 + (n % 64 > 0 ? 1 : 0);
}

uint64_t bitwiseTensorSize(const dim3 dim) {
	return numberOfUint64s(dim.x) * dim.y * dim.z;
}

uint64_t prod(const dim3 dim) {
	return dim.x * dim.y * dim.z;
}

/**
 * Allocates zeroed memory of type uint64_t.
 */
uint64_t* checkedCalloc(const uint64_t numberOfElements) {
	uint64_t* result = calloc(numberOfElements, sizeof(uint64_t));
	if (result == NULL ) {
		fprintf(stderr,
				"Failed to allocate space for %zu uint64_t values. Original message: %s\n",
				numberOfElements, strerror(errno));
		exit(EXIT_FAILURE);
	}
	return result;
}

/*
 * Prints the binary representation of 64 bits. The least significant bit
 * comes last.
 */
void printBinary(const uint64_t n) {
	for (int shift = sizeof(uint64_t) * 8 - 1; shift >= 0; shift--) {
		printf((n >> shift & 1) ? "1" : "0");
		printf((shift % 8 == 0) ? " " : "");
	}
	printf("\n");
}

/*
 * Prints a bitwise represented tensor.
 */
void printBitwiseTensor(const uint64_t* tensor, const dim3 dim) {
	const uint64_t integerPerColumn = numberOfUint64s(dim.x);
	for (uint64_t i = 0; i < bitwiseTensorSize(dim); i++) {
		printBinary(tensor[i]);
		if (i % integerPerColumn == integerPerColumn - 1)
			printf("\n");
		if (i % dim.y == dim.y - 1)
			printf("\n");
	}
	printf("\n");
}

/*
 * Prints a tensor slice by slice (as in Matlab).
 */
void printTensorRowWise(const uint64_t* tensor, const dim3 dim) {
	for (uint64_t k = 0; k < dim.z; k++) {
		for (uint64_t i = 0; i < dim.x; i++) {
			for (uint64_t j = 0; j < dim.y; j++) {
				printf("%lu", tensor[i + j * dim.x + k * dim.x * dim.y]);
			}
			printf("\n");
		}
		printf("\n\n");
	}
}

/*
 * Prints a tensor slice by slice with the columns as rows.
 */
void printTensorColWise(const bool* tensor, const dim3 dim) {
	for (uint64_t k = 0; k < dim.z; k++) {
		for (uint64_t j = 0; j < dim.y; j++) {
			uint64_t space = 0;
			uint64_t onesInCol = 0;
			for (uint64_t i = 0; i < dim.x; i++) {
				onesInCol += tensor[i + j * dim.x + k * dim.x * dim.y];
				printf("%d", tensor[i + j * dim.x + k * dim.x * dim.y]);
				if (space % 8 == 7)
					printf(" ");
				space++;
			}
			printf(" -- %lu\n", onesInCol);
		}
		printf("\n\n");
	}
}

/**
 * Prints a sparse binary 3-way tensor.
 */
void printSparseBinaryTensor(const double* data, const uint64_t numberOfRows) {
	for (uint64_t i = 0; i < numberOfRows; i++) {
		uint64_t c1 = (uint64_t) data[i];
		uint64_t c2 = (uint64_t) data[i + numberOfRows];
		uint64_t c3 = (uint64_t) data[i + numberOfRows + numberOfRows];
		printf("%lu %lu %lu \n", c1, c2, c3);
	}
}

/*
 * Obtains a bit representation from a binary tensor with x rows, y columns, and
 * z rows. The binary representation is column-based and the columns are padded
 * to 64 bits.
 * To save iterations, the number of ones in each column is provided together with
 * the tensor.
 */
bitwiseTensor* toBitRepresentation(const bool* matlabTensor, const dim3 dim) {
	/* determine size for bitwiseTensor */
	const uint64_t intPerCol = numberOfUint64s(dim.x);
	bitwiseTensor* result = malloc(sizeof(bitwiseTensor));
	result->tensor = checkedCalloc(bitwiseTensorSize(dim));
	result->onesPerColumn = checkedCalloc(dim.y * dim.z);

	for (uint64_t k = 0; k < dim.z; k++) {
		for (uint64_t j = 0; j < dim.y; j++) {
			for (uint64_t i = 0; i < dim.x; i++) {
				if (matlabTensor[i + j * dim.x + k * dim.x * dim.y] > 0) {
					result->tensor[k * intPerCol * dim.y + j * intPerCol + i / 64] |= (ONE << i % 64);
					result->onesPerColumn[j + k * dim.y]++;
				}
			}
		}
	}

	return result;
}

/*
 * Obtains a matlab-compatible representation of a bit-wise represented tensor.
 */
bool* toMatlabRepresentation(const uint64_t* bitwiseTensor, const dim3 dim) {
	const uint64_t intPerCol = numberOfUint64s(dim.x);
	bool* matlabTensor = calloc(dim.x * dim.y * dim.z, sizeof(bool));
	if (matlabTensor == NULL ) {
		fprintf(stderr, "Memory allocation failed.\n");
		exit(EXIT_FAILURE);
	}

	for (uint64_t k = 0; k < dim.z; k++) {
		for (uint64_t j = 0; j < dim.y; j++) {
			for (uint64_t i = 0; i < dim.x; i++) {
				if (bitwiseTensor[k * intPerCol * dim.y + j * intPerCol + i / 64] & (ONE << i % 64)) {
					matlabTensor[i + j * dim.x + k * dim.x * dim.y] = true;
				}
			}
		}
	}

	return matlabTensor;
}

bitwiseTensor* sparseToBitRepresentation(const double* coordinates, const uint64_t len, const dim3 dim) {
	const uint64_t intPerCol = numberOfUint64s(dim.x);
	bitwiseTensor* result = malloc(sizeof(bitwiseTensor));
	result->tensor = checkedCalloc(bitwiseTensorSize(dim));
	result->onesPerColumn = checkedCalloc(dim.y * dim.z);

	for (uint64_t i = 0; i < len; i++) {
		uint64_t c1 = (uint64_t) coordinates[i] - 1;
		uint64_t c2 = (uint64_t) coordinates[i + len] - 1;
		uint64_t c3 = (uint64_t) coordinates[i + len + len] - 1;
		result->tensor[c1 / 64 + c2 * intPerCol + c3 * intPerCol * dim.y] |= (ONE << c1 % 64);
		result->onesPerColumn[c2 + c3 * dim.y]++;
	}

	return result;
}

arrayWithLength* toSparseMatlabRepresentation(const uint64_t* bitwiseTensor, const dim3 dim) {
	const uint64_t intPerCol = numberOfUint64s(dim.x);
	uint64_t numCoords = 0;
	for (uint64_t i = 0; i < bitwiseTensorSize(dim); i++) {
		numCoords += POPCNT64(bitwiseTensor[i]);
	}
	uint64_t len = numCoords * 3;
	arrayWithLength* result;
	result = malloc(sizeof(arrayWithLength));
	result->length = len;
	result->data = calloc(len, sizeof(double));
	if (result->data == NULL ) {
		fprintf(stderr, "Memory allocation failed.\n");
		exit(EXIT_FAILURE);
	}

	uint64_t currentCoord = 0;
	for (uint64_t k = 0; k < dim.z; k++) {
		for (uint64_t j = 0; j < dim.y; j++) {
			for (uint64_t i = 0; i < dim.x; i++) {
				if (bitwiseTensor[j * intPerCol + i / 64] & (ONE << i % 64)) {
					result->data[currentCoord] = i + 1;
					result->data[currentCoord + numCoords] = j + 1;
					result->data[currentCoord + numCoords + numCoords] = k + 1;
					currentCoord++;
				}
			}
		}
	}

	return result;
}
